;;;-*-Mode:LISP; Package:(PCL (LISP WALKER)); Base:10; Syntax:Common-lisp -*-
;;;
;;; *************************************************************************
;;; Copyright (c) 1985, 1986, 1987, 1988, 1989, 1990 Xerox Corporation.
;;; All rights reserved.
;;;
;;; Use and copying of this software and preparation of derivative works
;;; based upon this software are permitted.  Any distribution of this
;;; software or derivative works must comply with all applicable United
;;; States export control laws.
;;; 
;;; This software is made available AS IS, and Xerox Corporation makes no
;;; warranty about the software, its performance or its conformity to any
;;; specification.
;;; 
;;; Any person obtaining a copy of this software is requested to send their
;;; name and post office or electronic mail address to:
;;;   CommonLoops Coordinator
;;;   Xerox PARC
;;;   3333 Coyote Hill Rd.
;;;   Palo Alto, CA 94304
;;; (or send Arpanet mail to CommonLoops-Coordinator.pa@Xerox.arpa)
;;;
;;; Suggestions, comments and requests for improvements are also welcome.
;;; *************************************************************************
;;;

(ext:file-comment
  "$Header: /home/gerd/cmucl/PCL-CVSROOT/pcl/env.lisp,v 1.14 2003/01/02 16:28:41 gerd Exp $")
;;;
;;; Basic environmental stuff.
;;;

(in-package :pcl)

;;;
;;;
;;;

;;; ANSI compliance wants default structure printer to use #S(...) format.
(defmethod print-object ((object structure-object) stream)
  (lisp::default-structure-print object stream 0))

(defgeneric describe-object (object stream))

(defmethod describe-object (object stream)
  (describe object stream))

(defmethod describe-object ((object slot-object) stream)
  (let* ((class (class-of object))
	 (slotds (slots-to-inspect class object))
	 (max-slot-name-length 0)
	 (instance-slotds ())
	 (class-slotds ())
	 (other-slotds ()))
    (flet ((adjust-slot-name-length (name)
	     (setq max-slot-name-length
		   (max max-slot-name-length
			(length (the string (symbol-name name))))))
	   (describe-slot (name value &optional (allocation () alloc-p))
	     (if alloc-p
		 (format stream
			 "~% ~A ~S ~VT  ~S"
			 name allocation (+ max-slot-name-length 7) value)
		 (format stream
			 "~% ~A~VT  ~S"
			 name max-slot-name-length value))))
      ;; Figure out a good width for the slot-name column.
      (dolist (slotd slotds)
	(adjust-slot-name-length (slot-definition-name slotd))
	(case (slot-definition-allocation slotd)
	  (:instance (push slotd instance-slotds))
	  (:class    (push slotd class-slotds))
	  (otherwise (push slotd other-slotds))))
      (setq max-slot-name-length  (min (+ max-slot-name-length 3) 30))
      (format stream "~%~S is an instance of class ~S:" object class)

      (when instance-slotds
	(format stream "~% The following slots have :INSTANCE allocation:")
	(dolist (slotd (nreverse instance-slotds))
	  (describe-slot (slot-definition-name slotd)
			 (slot-value-or-default object (slot-definition-name slotd)))))

      (when class-slotds
	(format stream "~% The following slots have :CLASS allocation:")
	(dolist (slotd (nreverse class-slotds))
	  (describe-slot (slot-definition-name slotd)
			 (slot-value-or-default object (slot-definition-name slotd)))))

      (when other-slotds 
	(format stream "~% The following slots have allocation as shown:")
	(dolist (slotd (nreverse other-slotds))
	  (describe-slot (slot-definition-name slotd)
			 (slot-value-or-default object (slot-definition-name slotd))
			 (slot-definition-allocation slotd))))
      (values))))

(defmethod slots-to-inspect ((class slot-class) (object slot-object))
  (class-slots class))

(defvar *describe-metaobjects-as-objects-p* nil)

(defmethod describe-object ((fun standard-generic-function) stream)
  (format stream "~A is a generic function.~%" fun)
  (format stream "Its arguments are:~%  ~S~%"
          (generic-function-pretty-arglist fun))
  (format stream "Its methods are:")
  (dolist (meth (generic-function-methods fun))
    (format stream "~2%    ~{~S ~}~:S =>~%"
            (method-qualifiers meth)
            (unparse-specializers meth))
    (describe-object (or (method-fast-function meth)
			 (method-function meth))
		     stream))
  (when *describe-metaobjects-as-objects-p*
    (call-next-method)))

;;;
;;;
;;;
(defmethod describe-object ((class class) stream)
  (flet ((pretty-class (c) (or (class-name c) c)))
    (macrolet ((ft (string &rest args) `(format stream ,string ,@args)))
      (ft "~&~S is a class, it is an instance of ~S.~%"
	  class (pretty-class (class-of class)))
      (let ((name (class-name class)))
	(if name
	    (if (eq class (find-class name nil))
		(ft "Its proper name is ~S.~%" name)
		(ft "Its name is ~S, but this is not a proper name.~%" name))
	    (ft "It has no name (the name is NIL).~%")))
      (ft "The direct superclasses are: ~:S, and the direct~%~
           subclasses are: ~:S.  The class precedence list is:~%~S~%~
           There are ~D methods specialized for this class."
	  (mapcar #'pretty-class (class-direct-superclasses class))
	  (mapcar #'pretty-class (class-direct-subclasses class))
	  (mapcar #'pretty-class (class-precedence-list class))
	  (length (specializer-direct-methods class)))))
  (when *describe-metaobjects-as-objects-p*
    (call-next-method)))

(defun describe-package (object stream)
  (unless (packagep object) (setq object (find-package object)))
  (format stream "~&~S is a ~S.~%" object (type-of object))
  (let ((nick (package-nicknames object)))
    (when nick
      (format stream "You can also call it~@[ ~{~S~^, ~} or~] ~S.~%"
	      (butlast nick) (first (last nick)))))  
  (let* ((internal (lisp::package-internal-symbols object))
	 (internal-count (- (lisp::package-hashtable-size internal)
				  (lisp::package-hashtable-free internal)))
	 (external (lisp::package-external-symbols object))
	 (external-count (- (lisp::package-hashtable-size external)
				  (lisp::package-hashtable-free external))))
    (format stream "It has ~D internal and ~D external symbols (~D total).~%"
	    internal-count external-count (+ internal-count external-count)))
  (let ((used (package-use-list object)))
    (when used
      (format stream "It uses the packages ~{~S~^, ~}.~%"
	      (mapcar #'package-name used))))
  (let ((users (package-used-by-list object)))
    (when users
      (format stream "It is used by the packages ~{~S~^, ~}.~%"
	      (mapcar #'package-name users)))))

(defmethod describe-object ((object package) stream)
  (describe-package object stream))

(defmethod describe-object ((object hash-table) stream)
  (format stream "~&~S is an ~a hash table."
	  object
	  (lisp::hash-table-test object))
  (format stream "~&Its size is ~d buckets."
	  (lisp::hash-table-size object))
  (format stream "~&Its rehash-size is ~d."
	  (lisp::hash-table-rehash-size object))
  (format stream "~&Its rehash-threshold is ~d."
	  (hash-table-rehash-threshold object))
  (format stream "~&It currently holds ~d entries."
	  (lisp::hash-table-number-entries object)))



;;; ***********************
;;; Method Tracing  *******
;;; ***********************

(defvar *traced-methods* ())

(defun %trace-method (spec &rest trace-options)
  "Trace method Spec using Trace options Trace-Options.
Spec can be a method object or a list
 (<method-name> <qualifier>* (<specializer>*)"
  (multiple-value-bind (gf method method-function-name)
      (parse-method-or-spec spec)
    (declare (ignore gf))
    (flet ((traced-method-function (method fast-function)
	     (let ((function (or fast-function (method-function method))))
	       (when (fboundp method-function-name)
		 (eval `(untrace ,method-function-name)))
	       (setf (symbol-function method-function-name) function)
	       (eval `(trace ,method-function-name ,@trace-options))
	       (symbol-function method-function-name))))
      ;;
      ;; Untrace the method if it's traced now.
      (when (assoc method *traced-methods*)
	(warn "~S is currently traced, retracing it." method)
	(%untrace-method spec))
      ;;
      ;; Set METHOD's fast-function or function to a traced version.
      ;; If it's a fast function, make sure that its plist is the same
      ;; as the original function's plist, otherwise the code computing
      ;; effective methods doesn't know how to call the traced function.
      (let* ((fast-function (method-fast-function method))
	     (new-function (traced-method-function method fast-function)))
	(cond (fast-function
	       (setf (slot-value method 'fast-function) new-function)
	       (setf (method-function-plist new-function)
		     (method-function-plist fast-function)))
	      (t
	       (setf (slot-value method 'function) new-function)))
	;;
	;; Record that method is traced.
	(assert (null (assoc method *traced-methods*)))
	(push (list method method-function-name
		    (if fast-function 'fast-function 'function))
	      *traced-methods*)
	;;
	;; Let effective methods be recomputed.
	(update-effective-methods method)))))

(defun update-effective-methods (method)
  (update-dfun (method-generic-function method)))

(defun remove-trace-entry (method)
  (let ((traced (assoc method *traced-methods*)))
    (when traced
      (setq *traced-methods* (delete traced *traced-methods*)))
    traced))
  
(defun %%untrace-method (traced)
  (destructuring-bind (method fn-name function-type) traced
    (when (eq function-type 'fast-function)
      (setf (method-function-plist (symbol-function fn-name)) nil))
    (eval `(untrace ,fn-name))
    (setf (slot-value method function-type) (symbol-function fn-name))
    (update-effective-methods method)))

(defun %untrace-method (spec)
  "Untrace the method specified by Spec.
Spec can be either a method object or a list of the form
  (<method-name> <qualifier>* (<specializer>*))."
  (multiple-value-bind (gf method method-function-name)
      (parse-method-or-spec spec)
    (declare (ignore gf method-function-name))
    (let ((traced (remove-trace-entry method)))
      (if traced
	  (%%untrace-method traced)
	  (warn "~S is not a traced method" method)))))

(defmacro trace-method (spec &rest trace-options)
  "Trace the method specified by Spec using Trace options Trace-Options.
Spec is a list of the form (<method-name> <qualifier>* (<specializer>*))."
  `(apply #'%trace-method ',spec ',trace-options))

(defmacro untrace-method (spec)
  "Untrace the method specified by Spec.
Spec is a list of the form (<method-name> <qualifier>* (<specializer>*))."
  `(%untrace-method ',spec))

(defun untrace-all-methods ()
  "Untrace all traced methods."
  (mapc #'%%untrace-method *traced-methods*)
  (setq *traced-methods* ()))

(defun trace-generic-function-methods (gf &optional trace-options)
  "Trace all methods of generic function Gf.
If Gf is a symbol, use its function value."
  (loop with gf = (if (symbolp gf) (gdefinition gf) gf)
	for method in (generic-function-methods gf) do
	  (apply #'%trace-method method trace-options)))
  
(defun untrace-generic-function-methods (gf)
  "Untrace all methods of generic function Gf.
If Gf is a symbol, use its function value."
  (loop with gf = (if (symbolp gf) (gdefinition gf) gf)
	for method in (generic-function-methods gf) do
	  (%untrace-method method)))
  

(defmacro undefmethod (&rest args)
  `(undefmethod-1 ',args))

(defun undefmethod-1 (args)
  (multiple-value-bind (gf method)
      (parse-method-or-spec args)
    (when (and gf method)
      (remove-method gf method)
      method)))


(pushnew :pcl *features*)
(pushnew :portable-commonloops *features*)
(pushnew :pcl-structures *features*)
(pushnew :gerds-pcl *features*)

(when (find-package "OLD-PCL")
  (setf (symbol-function (find-symbol "PRINT-OBJECT" :old-pcl))
        (symbol-function 'pcl::print-object)))


;;;; MAKE-LOAD-FORM

(export '(cl::make-load-form cl::make-load-form-saving-slots) "CL")

(defgeneric make-load-form (object &optional environment))

(defmethod make-load-form ((object structure-object) &optional environment)
  (declare (ignore environment))
  (kernel:make-structure-load-form object))

(defmethod make-load-form ((object wrapper) &optional env)
  (declare (ignore env))
  (let ((pname (kernel:class-proper-name (kernel:layout-class object))))
    (unless pname
      (error "Can't dump wrapper for anonymous class:~%  ~S"
	     (kernel:layout-class object)))
    `(kernel:class-layout (lisp:find-class ',pname))))

(defun make-load-form-saving-slots (object &key slot-names environment)
  (declare (ignore environment))
  (when slot-names
    (warn ":SLOT-NAMES MAKE-LOAD-FORM option not implemented, dumping all ~
	   slots:~%  ~S"
	  object))
  :just-dump-it-normally)


;;; The following are hacks to deal with CMU CL having two different CLASS
;;; classes.
;;;
(defun coerce-to-pcl-class (class)
  (if (typep class 'lisp:class)
      (or (kernel:class-pcl-class class)
	  (find-structure-class (lisp:class-name class)))
      class))

(defmethod make-instance ((class lisp:class) &rest stuff)
  (apply #'make-instance (coerce-to-pcl-class class) stuff))
(defmethod change-class (instance (class lisp:class) &rest initargs)
  (apply #'change-class instance (coerce-to-pcl-class class) initargs))

(macrolet ((frob (&rest names)
	     `(progn
		,@(mapcar (lambda (name)
			    `(defmethod ,name ((class lisp:class))
			      (funcall #',name
			       (coerce-to-pcl-class class))))
			  names))))
  (frob
    class-direct-slots
    class-prototype
    class-precedence-list
    class-direct-default-initargs
    class-direct-superclasses
    compute-class-precedence-list
    class-default-initargs class-finalized-p
    class-direct-subclasses class-slots
    make-instances-obsolete))
